@@ -54,6 +54,25 @@ def _get_app_and_auth_code(
     assert ac is not None
     return (app, ac, redirect_uri)
 
+def _render(url, description=None):
+    # Render a url in html if description is available, otherwise return url as-is
+    return "<a href='{url}' target=_blank>{description}</a>".format(
+        url=url, description=description) if description else url
+
+
+def _get_hint(html_mode=None, username=None, lab_name=None, username_uri=None):
+    return "Sign in with {user} whose password is available from {lab}".format(
+        user=("<b>{}</b>".format(username) if html_mode else username)
+            if username
+            else "the upn from {}".format(_render(
+                username_uri, description="here" if html_mode else None)),
+        lab=_render(
+            "https://aka.ms/GetLabUserSecret?Secret=" + (lab_name or "msidlabXYZ"),
+            description="this password api" if html_mode else None,
+            ),
+        )
+
+
 @unittest.skipIf(os.getenv("TRAVIS_TAG"), "Skip e2e tests during tagged release")
 class E2eTestCase(unittest.TestCase):
 
@@ -86,7 +105,7 @@ class E2eTestCase(unittest.TestCase):
         self.assertNotEqual(0, len(accounts))
         account = accounts[0]
         if ("scope" not in result_from_wire  # This is the usual case
-                or  # Authority server could reject some scopes
+                or  # Authority server could return different set of scopes
                 set(scope) <= set(result_from_wire["scope"].split(" "))
                 ):
             # Going to test acquire_token_silent(...) to locate an AT from cache
@@ -115,7 +134,7 @@ class E2eTestCase(unittest.TestCase):
             #   result_from_wire['access_token'] != result_from_cache['access_token']
             # but ROPC in B2C tends to return the same AT we obtained seconds ago.
             # Now looking back, "refresh_token grant would return a brand new AT"
-            # was just an empirical observation but never a committment in specs,
+            # was just an empirical observation but never a commitment in specs,
             # so we adjust our way to assert here.
             (result_from_cache or {}).get("access_token"),
             "We should get an AT from acquire_token_silent(...) call")
@@ -133,6 +152,36 @@ class E2eTestCase(unittest.TestCase):
         self.assertEqual(
             result_from_wire['access_token'], result_from_cache['access_token'],
             "We should get a cached AT")
+        self.app.acquire_token_silent(
+            # Result will typically be None, because client credential grant returns no RT.
+            # But we care more on this call should succeed without exception.
+            scope, account=None,
+            force_refresh=True)  # Mimic the AT already expires
+
+    @classmethod
+    def _build_app(cls,
+            client_id,
+            client_credential=None,
+            authority="https://login.microsoftonline.com/common",
+            scopes=["https://graph.microsoft.com/.default"],  # Microsoft Graph
+            http_client=None,
+            azure_region=None,
+            **kwargs):
+        try:
+            import pymsalruntime
+            broker_available = True
+        except ImportError:
+            broker_available = False
+        return (msal.ConfidentialClientApplication
+                if client_credential else msal.PublicClientApplication)(
+            client_id,
+            client_credential=client_credential,
+            authority=authority,
+            azure_region=azure_region,
+            http_client=http_client or MinimalHttpClient(),
+            allow_broker=broker_available  # This way, we reuse same test cases, by run them with and without broker
+                and not client_credential,
+            )
 
     def _test_username_password(self,
             authority=None, client_id=None, username=None, password=None, scope=None,
@@ -141,9 +190,9 @@ class E2eTestCase(unittest.TestCase):
             http_client=None,
             **ignored):
         assert authority and client_id and username and password and scope
-        self.app = msal.ClientApplication(
+        self.app = self._build_app(
             client_id, authority=authority,
-            http_client=http_client or MinimalHttpClient(),
+            http_client=http_client,
             azure_region=azure_region,  # Regional endpoint does not support ROPC.
                 # Here we just use it to test a regional app won't break ROPC.
             client_credential=client_secret)
@@ -158,8 +207,7 @@ class E2eTestCase(unittest.TestCase):
     def _test_device_flow(
             self, client_id=None, authority=None, scope=None, **ignored):
         assert client_id and authority and scope
-        self.app = msal.PublicClientApplication(
-            client_id, authority=authority, http_client=MinimalHttpClient())
+        self.app = self._build_app(client_id, authority=authority)
         flow = self.app.initiate_device_flow(scopes=scope)
         assert "user_code" in flow, "DF does not seem to be provisioned: %s".format(
             json.dumps(flow, indent=4))
@@ -183,23 +231,31 @@ class E2eTestCase(unittest.TestCase):
 
     def _test_acquire_token_interactive(
             self, client_id=None, authority=None, scope=None, port=None,
-            username_uri="",  # But you would want to provide one
+            username=None, lab_name=None,
+            username_uri="",  # Unnecessary if you provided username and lab_name
             data=None,  # Needed by ssh-cert feature
+            prompt=None,
+            enable_msa_passthrough=None,
             **ignored):
         assert client_id and authority and scope
-        self.app = msal.PublicClientApplication(
-            client_id, authority=authority, http_client=MinimalHttpClient())
+        self.app = self._build_app(client_id, authority=authority)
+        logger.info(_get_hint(  # Useful when testing broker which shows no welcome_template
+            username=username, lab_name=lab_name, username_uri=username_uri))
         result = self.app.acquire_token_interactive(
             scope,
+            login_hint=username,
+            prompt=prompt,
             timeout=120,
             port=port,
+            parent_window_handle=self.app.CONSOLE_WINDOW_HANDLE,  # This test app is a console app
+            enable_msa_passthrough=enable_msa_passthrough,  # Needed when testing MSA-PT app
             welcome_template=  # This is an undocumented feature for testing
                 """<html><body><h1>{id}</h1><ol>
-    <li>Get a username from the upn shown at <a href="{username_uri}">here</a></li>
-    <li>Get its password from https://aka.ms/GetLabUserSecret?Secret=msidlabXYZ
-        (replace the lab name with the labName from the link above).</li>
+    <li>{hint}</li>
     <li><a href="$auth_uri">Sign In</a> or <a href="$abort_uri">Abort</a></li>
-    </ol></body></html>""".format(id=self.id(), username_uri=username_uri),
+    </ol></body></html>""".format(id=self.id(), hint=_get_hint(
+                html_mode=True,
+                username=username, lab_name=lab_name, username_uri=username_uri)),
             data=data or {},
             )
         self.assertIn(
@@ -208,6 +264,11 @@ class E2eTestCase(unittest.TestCase):
                 # Note: No interpolation here, cause error won't always present
                 error=result.get("error"),
                 error_description=result.get("error_description")))
+        if username and result.get("id_token_claims", {}).get("preferred_username"):
+            self.assertEqual(
+                username, result["id_token_claims"]["preferred_username"],
+                "You are expected to sign in as account {}, but tokens returned is for {}".format(
+                    username, result["id_token_claims"]["preferred_username"]))
         self.assertCacheWorksForUser(result, scope, username=None, data=data or {})
         return result  # For further testing
 
@@ -229,7 +290,7 @@ class SshCertTestCase(E2eTestCase):
         self.assertEqual("ssh-cert", result["token_type"])
 
     @unittest.skipIf(os.getenv("TRAVIS"), "Browser automation is not yet implemented")
-    def test_ssh_cert_for_user(self):
+    def test_ssh_cert_for_user_should_work_with_any_account(self):
         result = self._test_acquire_token_interactive(
             client_id="04b07795-8ddb-461a-bbee-02f9e1bf7b46",  # Azure CLI is one
                 # of the only 2 clients that are PreAuthz to use ssh cert feature
@@ -237,6 +298,7 @@ class SshCertTestCase(E2eTestCase):
             scope=self.SCOPE,
             data=self.DATA1,
             username_uri="https://msidlab.com/api/user?usertype=cloud",
+            prompt="none" if msal.application._is_running_in_cloud_shell() else None,
             )   # It already tests reading AT from cache, and using RT to refresh
                 # acquire_token_silent() would work because we pass in the same key
         self.assertIsNotNone(result.get("access_token"), "Encountered {}: {}".format(
@@ -254,6 +316,20 @@ class SshCertTestCase(E2eTestCase):
         self.assertNotEqual(result["access_token"], refreshed_ssh_cert['access_token'])
 
 
+@unittest.skipUnless(
+    msal.application._is_running_in_cloud_shell(),
+    "Manually run this test case from inside Cloud Shell")
+class CloudShellTestCase(E2eTestCase):
+    app = msal.PublicClientApplication("client_id")
+    scope_that_requires_no_managed_device = "https://management.core.windows.net/"  # Scopes came from https://msazure.visualstudio.com/One/_git/compute-CloudShell?path=/src/images/agent/env/envconfig.PROD.json&version=GBmaster&_a=contents
+    def test_access_token_should_be_obtained_for_a_supported_scope(self):
+        result = self.app.acquire_token_interactive(
+            [self.scope_that_requires_no_managed_device], prompt="none")
+        self.assertEqual(
+            "Bearer", result.get("token_type"), "Unexpected result: %s" % result)
+        self.assertIsNotNone(result.get("access_token"))
+
+
 THIS_FOLDER = os.path.dirname(__file__)
 CONFIG = os.path.join(THIS_FOLDER, "config.json")
 @unittest.skipUnless(os.path.exists(CONFIG), "Optional %s not found" % CONFIG)
@@ -509,7 +585,8 @@ class LabBasedTestCase(E2eTestCase):
 
     def _test_acquire_token_by_auth_code_flow(
             self, client_id=None, authority=None, port=None, scope=None,
-            username_uri="",  # But you would want to provide one
+            username=None, lab_name=None,
+            username_uri="",  # Optional if you provided username and lab_name
             **ignored):
         assert client_id and authority and scope
         self.app = msal.ClientApplication(
@@ -522,11 +599,11 @@ class LabBasedTestCase(E2eTestCase):
             auth_response = receiver.get_auth_response(
                 auth_uri=flow["auth_uri"], state=flow["state"], timeout=60,
                 welcome_template="""<html><body><h1>{id}</h1><ol>
-    <li>Get a username from the upn shown at <a href="{username_uri}">here</a></li>
-    <li>Get its password from https://aka.ms/GetLabUserSecret?Secret=msidlabXYZ
-        (replace the lab name with the labName from the link above).</li>
+    <li>{hint}</li>
     <li><a href="$auth_uri">Sign In</a> or <a href="$abort_uri">Abort</a></li>
-    </ol></body></html>""".format(id=self.id(), username_uri=username_uri),
+    </ol></body></html>""".format(id=self.id(), hint=_get_hint(
+                    html_mode=True,
+                    username=username, lab_name=lab_name, username_uri=username_uri)),
                 )
         if auth_response is None:
             self.skipTest("Timed out. Did not have test settings in hand? Prepare and retry.")
@@ -546,6 +623,11 @@ class LabBasedTestCase(E2eTestCase):
                 # Note: No interpolation here, cause error won't always present
                 error=result.get("error"),
                 error_description=result.get("error_description")))
+        if username and result.get("id_token_claims", {}).get("preferred_username"):
+            self.assertEqual(
+                username, result["id_token_claims"]["preferred_username"],
+                "You are expected to sign in as account {}, but tokens returned is for {}".format(
+                    username, result["id_token_claims"]["preferred_username"]))
         self.assertCacheWorksForUser(result, scope, username=None)
 
     def _test_acquire_token_obo(self, config_pca, config_cca,
@@ -601,11 +683,12 @@ class LabBasedTestCase(E2eTestCase):
             self, client_id=None, client_secret=None, authority=None, scope=None,
             **ignored):
         assert client_id and client_secret and authority and scope
-        app = msal.ConfidentialClientApplication(
+        self.app = msal.ConfidentialClientApplication(
             client_id, client_credential=client_secret, authority=authority,
             http_client=MinimalHttpClient())
-        result = app.acquire_token_for_client(scope)
+        result = self.app.acquire_token_for_client(scope)
         self.assertIsNotNone(result.get("access_token"), "Got %s instead" % result)
+        self.assertCacheWorksForApp(result, scope)
 
 
 class WorldWideTestCase(LabBasedTestCase):
@@ -642,10 +725,23 @@ class WorldWideTestCase(LabBasedTestCase):
 
     @unittest.skipIf(os.getenv("TRAVIS"), "Browser automation is not yet implemented")
     def test_cloud_acquire_token_interactive(self):
-        config = self.get_lab_user(usertype="cloud")
-        self._test_acquire_token_interactive(
-            username_uri="https://msidlab.com/api/user?usertype=cloud",
-            **config)
+        self._test_acquire_token_interactive(**self.get_lab_user(usertype="cloud"))
+
+    @unittest.skipIf(os.getenv("TRAVIS"), "Browser automation is not yet implemented")
+    def test_msa_pt_app_signin_via_organizations_authority_without_login_hint(self):
+        """There is/was an upstream bug. See test case full docstring for the details.
+
+        When a MSAL-PT flow that account control is launched, user has 2+ AAD accounts in WAM,
+        selects an AAD account that is NOT the default AAD account from the OS,
+        it will incorrectly get tokens for default AAD account.
+        """
+        self._test_acquire_token_interactive(**dict(
+            self.get_lab_user(usertype="cloud"),  # This is generally not the current laptop's default AAD account
+            authority="https://login.microsoftonline.com/organizations",
+            client_id="04b07795-8ddb-461a-bbee-02f9e1bf7b46",  # Azure CLI is an MSA-PT app
+            enable_msa_passthrough=True,
+            prompt="select_account",  # In MSAL Python, this resets login_hint
+            ))
 
     def test_ropc_adfs2019_onprem(self):
         # Configuration is derived from https://github.com/AzureAD/microsoft-authentication-library-for-dotnet/blob/4.7.0/tests/Microsoft.Identity.Test.Common/TestConstants.cs#L250-L259
@@ -672,22 +768,22 @@ class WorldWideTestCase(LabBasedTestCase):
     @unittest.skipIf(os.getenv("TRAVIS"), "Browser automation is not yet implemented")
     def test_adfs2019_onprem_acquire_token_by_auth_code_flow(self):
         config = self.get_lab_user(usertype="onprem", federationProvider="ADFSv2019")
-        config["authority"] = "https://fs.%s.com/adfs" % config["lab_name"]
-        config["scope"] = self.adfs2019_scopes
-        config["port"] = 8080
-        self._test_acquire_token_by_auth_code_flow(
-            username_uri="https://msidlab.com/api/user?usertype=onprem&federationprovider=ADFSv2019",
-            **config)
+        self._test_acquire_token_by_auth_code_flow(**dict(
+            config,
+            authority="https://fs.%s.com/adfs" % config["lab_name"],
+            scope=self.adfs2019_scopes,
+            port=8080,
+            ))
 
     @unittest.skipIf(os.getenv("TRAVIS"), "Browser automation is not yet implemented")
     def test_adfs2019_onprem_acquire_token_interactive(self):
         config = self.get_lab_user(usertype="onprem", federationProvider="ADFSv2019")
-        config["authority"] = "https://fs.%s.com/adfs" % config["lab_name"]
-        config["scope"] = self.adfs2019_scopes
-        config["port"] = 8080
-        self._test_acquire_token_interactive(
-            username_uri="https://msidlab.com/api/user?usertype=onprem&federationprovider=ADFSv2019",
-            **config)
+        self._test_acquire_token_interactive(**dict(
+            config,
+            authority="https://fs.%s.com/adfs" % config["lab_name"],
+            scope=self.adfs2019_scopes,
+            port=8080,
+            ))
 
     @unittest.skipUnless(
         os.getenv("LAB_OBO_CLIENT_SECRET"),
@@ -769,14 +865,12 @@ class WorldWideTestCase(LabBasedTestCase):
 
     @unittest.skipIf(os.getenv("TRAVIS"), "Browser automation is not yet implemented")
     def test_b2c_acquire_token_by_auth_code_flow(self):
-        config = self.get_lab_app_object(azureenvironment="azureb2ccloud")
-        self._test_acquire_token_by_auth_code_flow(
+        self._test_acquire_token_by_auth_code_flow(**dict(
+            self.get_lab_user(usertype="b2c", b2cprovider="local"),
             authority=self._build_b2c_authority("B2C_1_SignInPolicy"),
-            client_id=config["appId"],
             port=3843,  # Lab defines 4 of them: [3843, 4584, 4843, 60000]
-            scope=config["scopes"],
-            username_uri="https://msidlab.com/api/user?usertype=b2c&b2cprovider=local",
-            )
+            scope=self.get_lab_app_object(azureenvironment="azureb2ccloud")["scopes"],
+            ))
 
     def test_b2c_acquire_token_by_ropc(self):
         config = self.get_lab_app_object(azureenvironment="azureb2ccloud")
